/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
#include "fm.h"
#include <ftw.h>
#include <string.h>

static char SccsID[] = "@(#)vtree.c	1.2\t11/25/92";            /* sccs keyword */

static char msg[200];

treeWindow *curTreeWnd;
treeWindow *makeTreeWindow();
dirNode *makeDirNodeStruct();

 /*
  VTREE
  
  This module has been implemented on the base of an application provided by:

  Gene Lee  UUCP: ...ihnp4!{meccts,dayton,rosevax}!ems!minnow!lee
  UNISYS Corporation     ATT:  (612) 635-6334
 */

static char *table[MAXCOL];   	/* holds the path - fragments */
static tablecount;	 	/* count of how many entries in the table
				   are currently valid */
static column;  		/* what column are we printing output in */
static match_found;  		/* has a match been found yet */
static firsttime;  		/* the first time this routine is called */

static currentRow, conCnt;
static dirNode *curDirNode;
static dirNode *father[MAXCOL];
static maxrow, maxcol;
static dirs_found;		/* number of directories found */

static int fn();

void buildFileTree(wnd)
     dirWindow *wnd;
{
  int rc, cnt;
  Position x, y;
  Arg args[4];
  treeWindow *tw;

  dirs_found = 0;
  curDirNode = (dirNode *) NULL;

  if (wnd->tree != (treeWindow *) NULL) {
    XMapRaised(XtDisplay(Head->root), XtWindow(wnd->tree->shell));
    XWarpPointer(XtDisplay(Head->root), None, XtWindow(wnd->tree->w),
	       0, 0, 0, 0, 0, 0);
    return;
  }
  currentRow = 0;
  conCnt = 0;
  maxrow = maxcol = 1;
  tablecount = 0;
  column = 0;
  firsttime = 1;
  
  initTable(table, MAXCOL);
  
  tw = makeTreeWindow(wnd); 
  curTreeWnd = tw;

  if ((rc = ftw(wnd->path, fn, 10)) < 0 && rc != -3) {
    exAlert(Head->root, 100, 100, "FTW error in vtree");
    deleteTreeWindow(wnd);
    freeTable(table, MAXCOL);
    return;
  }
  
  if (! dirs_found) {			/* we found no directories */
    freeTable(table, MAXCOL);
    free(tw->path);
    free(tw);
    sprintf(msg, "No directories in\n-- %s -- !", wnd->path);
    exAlert(Head->root, 100, 100, msg);
    return;
  }

  tw->rows = maxrow;
  tw->cols = maxcol;
  tw->conCnt = conCnt;
  
  cnt = 0;
  XtSetArg(args[cnt], XmNx, &x); cnt++;
  XtSetArg(args[cnt], XmNy, &y); cnt++;
  XtGetValues(wnd->fmTop, args, cnt);

  createDirNodes(tw);

  XtPopup(curTreeWnd->shell, XtGrabNone);

  freeTable(table, MAXCOL);
  wnd->tree = curTreeWnd;
}

/* ARGSUSED */
static int fn(name, ptr, type)
     char *name;          	/* name of the directory entity */
     struct stat *ptr;    	/* pointer to a STAT(2) structure */
     int type;            	/* entity type according to <ftw.h> */
{
  if (type == FTW_DNR) {
    sprintf(msg, "Vtree: directory can't be read: %s", name);
    exAlert(Head->root, 100, 100, msg);
    return(0);
  } else if (type != FTW_D) 
    return(0);  		/* other types of entity need no processing */
 
  if (firsttime) {
    loadtable(table, &tablecount, name);
    firsttime = 0;
    return(0);
  }

  dirs_found++;

  match_found=0;
  do {
    /* if this path is an extension of the last path */
    if (tablecount == 0 || match(table, tablecount, name) == 0) {
      tableInsert(table, tablecount, tail(name));
      tablecount++;
      column++;  	/* were going to print new name in next column */
      if (drawlevel(curTreeWnd, column, name) < 0) {
	exAlert(Head->root, 100, 100, "Too many nodes: running out of memory!");
	return(-3);
      }
      
      match_found = 1;  /* we just found a new directory level */
    } else {
      tablecount--;  		/* back up one node level and try again */
      column--;			/* printout column goes back with dir level */
    }
    
    if (tablecount < 0 || tablecount >= MAXCOL) {
      exAlert(Head->root, 100, 100, "Too many levels: running out of memory!");
      return(-3);
    }
  } while (!match_found);
  
  return(0);
}

/* match: *********************************************************************
 * match checks to see if the first nodes in the path name match to 
 * node names in the table
 *****************************************************************************/

int match(tab, tabcnt, name)
     char *tab[MAXCOL];
     int tabcnt;    	        /* how many entries in the table are valid 
				   and need to be checked */
     char *name;        	/* the directory path of node names */
{
  register i;
  char path[200];   		/* make a copy of pathname to strtok on */
  
  strcpy(path, name);
  
  /* check first node name against first table entry */
  if (strcmp(tab[0], strtok(path, "/")) != 0)
    return(-1);
  
  /* check the rest of the table up to the last valid entry against
     each node name in the path */
  for (i = 1; i < tabcnt; i++) 
    if (strcmp(tab[i], strtok(NULL,"/")) != 0)
      return(-1);  
  
  return(0);
}

/* loadtable: ****************************************************************
 * 
 *****************************************************************************/

loadtable(tab, tabcnt, name)
     char tab[MAXCOL][NODELEN];
     int *tabcnt;
     char *name;
{
  char path[160];   /* make a copy of the path so as not to chg it */
  char *ptr;
  
  strcpy(path, name);
  
  *tabcnt = 0;	   	/* nothing in the table yet */
  
  if ((ptr = strtok(path, "/")) == NULL)
    return;
  else {
    tableInsert(tab, *tabcnt, ptr);
    (*tabcnt)++;  		/* add the node name found to the count */
  }
  
  while ((ptr = strtok(NULL, "/")) != NULL) { /* while a node name is
						 found */
    tableInsert(tab, *tabcnt, ptr);
    (*tabcnt)++;
  }
}

/* drawlevel: ****************************************************************
 * 
 *****************************************************************************/

drawlevel(tw, col, path)
     treeWindow *tw;
     int col;                           /* column to print nodename in */
     char *path;                        /* file name (absolute) */
{
  static int lastcolumn = MAXCOL;  	/* last column a name was printed in */
  if (col >= MAXCOL || currentRow >= MAXROW) 
    return(-3);
  
  if (col <= lastcolumn)    		/* we need to start a new row */
    startnewrow(tw, col, path);
  else   				/* add name to end of last row */
    addtorow(tw, col, path);
  
  lastcolumn = col;      		/* remember last col we printed in */
  return(0);
}

/* addtorow: ****************************************************************
 * 
 *****************************************************************************/

addtorow(tw, col, path)
     treeWindow *tw;
     int col;	        		/* column where button should appear */
     char *path;
{
  dirNode *d;
  d = makeDirNodeStruct(tw, col, currentRow, path);
  connectDirNodes(tw, curDirNode, d);
  father[col] = curDirNode;
  curDirNode = d;
  maxcol = MAX(maxcol, col);
}

/* startnewrow: **************************************************************
 * 
 *****************************************************************************/

startnewrow(tw, col, path)
     treeWindow *tw;
     int col;   			/* column to print nodename in */
     char *path;
{
  dirNode *d;
  
  currentRow++;
  
  d = makeDirNodeStruct(tw, col, currentRow, path);
  connectDirNodes(tw, father[col], d);
  father[col + 1] = d;
  curDirNode = d;
  maxrow = MAX(maxrow, currentRow);
}

/* connectDirNodes: **********************************************************
 * 
 *****************************************************************************/

connectDirNodes(tw, from, to)
     treeWindow *tw;
     dirNode *from, *to;
{
  
  if ((! from) || (! to))
    return;
  
  tw->connections[conCnt].x1 = from->x + DIRBUTTONWIDTH(tw->font);
  tw->connections[conCnt].y1 = from->y + DIRBUTTONHEIGHT(tw->font) / 2;
  tw->connections[conCnt].x2 = to->x;
  tw->connections[conCnt].y2 = to->y + DIRBUTTONHEIGHT(tw->font) / 2;
  
  conCnt++;
}

  
