/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
/****************************************************************************
 * This file contains the functions generating the history tracer and 
 * handling the input/output of history tracer (main event loop); the 
 * functions corresponding to the menu options are also defined in here 
 ****************************************************************************/

#include <stdio.h>
#include <sys/file.h>
#include <Xm/ScrolledW.h>
#include <Xm/ScrollBar.h>
#include <Xm/SeparatoG.h>
#include <Xm/Form.h>
#include <exMenu.h>
#include "HT.h"
#include "histCmds.h"
#include <icons/hist>
#include "version.h"

extern char *exGetFilename();

static void HEdoScroll();
static void initHist();
static void getHistMenuResult();

int quitHistory(/* parms */);
int restoreHistory(/* parms */);
int saveHistory(/* parms */);
int stopHistory(/* parms */);
int contHistory(/* parms */);

static processEvents();

historyWindow *exHistory;
XFontStruct *TitleFont;
MenuDescr mWins[HISTMENUCNT];
commandStatistics *comStat;

int STOP = 0;            /* if set commands are not saved in history tracer */

/* EditHistory **************************************************************
 * initializes windows and editor; contains event loop
 ****************************************************************************/

EditHistory(parent, x, y, l, c)
     Widget parent;
     int x, y,                /* position of father window */
         l, c;                /* size of editor window in lines and columns */
{
  Arg args[20];
  register cnt;
  Widget neWidget, sep;
  Display *display;

  int DELETED = 1,
      histMenuCnt, 
      histHeight, histWidth,
      fw, fh,
      nfds, maxfd = getdtablesize();
  fd_set readfds[2];
  static char fullName[256];
  int dpy, scrlbWidth; 
  XWMHints wm_hints;
  XSizeHints size_hints;
  XtTranslations ptt; 

  initHist();
 
  display = XtDisplay(parent);

  exHistory = (historyWindow *) Malloc(sizeof(historyWindow));

  cnt = 0;
  XtSetArg(args[cnt], XmNx, x); cnt++;
  XtSetArg(args[cnt], XmNy, y); cnt++;
  XtSetArg(args[cnt], XmNtitle, VERSION); cnt++;
  XtSetArg(args[cnt], XmNiconName, "ERGO History Tracer"); cnt++;
  XtSetArg(args[cnt], XmNiconX, 1060); cnt++;
  XtSetArg(args[cnt], XmNiconY, 500); cnt++;
  XtSetArg(args[cnt], XmNkeyboardFocusPolicy, XmPOINTER); cnt++;

  exHistory->histShell = XtCreatePopupShell(VERSION, 
			 applicationShellWidgetClass, parent, args, cnt);

  /*** create widgets of history tracer ***/

  cnt = 0;
  XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;
  XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;

  exHistory->histForm = XtCreateManagedWidget("histForm", 
					      xmFormWidgetClass, 
					     exHistory->histShell, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;

  exHistory->menuBar = exCreateMenuBar(exHistory->histForm, args,
				       cnt, getHistMenuResult);

  makeFullName(getMenuDir(), HISTMENUS, fullName, sizeof(fullName));

  histMenuCnt = 1;
  initMenus(exHistory->menuBar, exHistory->histShell, fullName, mWins, 
	    &histMenuCnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNshadowThickness, 0); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, 1); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget,(Widget)exHistory->menuBar->menubar); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  /*XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;*/

  exHistory->scrollbar = XtCreateManagedWidget("histScrollB", 
					       xmScrollBarWidgetClass, 
					      exHistory->histForm, args, cnt); 

  XtAddCallback(exHistory->scrollbar, XmNincrementCallback,     HEdoScroll, 
		exHistory);
  XtAddCallback(exHistory->scrollbar, XmNdecrementCallback,     HEdoScroll, 
		exHistory);
  XtAddCallback(exHistory->scrollbar, XmNdragCallback,          HEdoScroll, 
		exHistory);
  XtAddCallback(exHistory->scrollbar, XmNpageDecrementCallback, HEdoScroll, 
		exHistory);
  XtAddCallback(exHistory->scrollbar, XmNpageIncrementCallback, HEdoScroll, 
		exHistory);

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget,(Widget)exHistory->menuBar->menubar); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, (Widget)exHistory->scrollbar); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  /* XtSetArg(args[cnt], XmNborderWidth, H_BORDER); cnt++;*/

  numDescr.lines = l;

  exHistory->numEdit = 
    makeEditor(exHistory->histForm, "numEdit", &numDescr, args, cnt);
  initCmdTables(exHistory->numEdit, histEditCmdTable, histEditEscCmds, 
		histCtrlXCmds, htCmds); 

  switchCursor(exHistory->numEdit, False);
  setViewMode(exHistory->numEdit);

  neWidget = (bufinfoptr(exHistory->numEdit))->w;
  cnt = 0;
  XtSetArg(args[cnt], XmNorientation, XmVERTICAL); cnt++;
  XtSetArg(args[cnt], XmNseparatorType, XmSINGLE_LINE); cnt++;
  XtSetArg(args[cnt], XmNshadowThickness, 0); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget,(Widget)exHistory->menuBar->menubar); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, (Widget)neWidget); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;

  sep = XtCreateManagedWidget("histGadget", xmSeparatorGadgetClass, 
			      exHistory->histForm, args, cnt); 

  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNtopWidget,(Widget)exHistory->menuBar->menubar); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, (Widget)sep); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  /* XtSetArg(args[cnt], XmNborderWidth, H_BORDER); cnt++;*/

  HEditDescr.lines = l;
  HEditDescr.cols = c;

  exHistory->comEdit = 
    makeEditor(exHistory->histForm, "comEdit", &HEditDescr, args, cnt);
  initCmdTables(exHistory->comEdit, histEditCmdTable, histEditEscCmds, 
		histCtrlXCmds, htCmds);
  XtAddActions(dialogCmds, 6); 
  ptt = XtParseTranslationTable(dialogCmdTable);
  XtOverrideTranslations((bufinfoptr(exHistory->comEdit))->edt, ptt);

  scrlbWidth = exFontHeight(getEditorFont(exHistory->comEdit)) * 3 / 2;
  cnt = 0;
  XtSetArg(args[cnt], XmNwidth, scrlbWidth); cnt++;
  XtSetArg(args[cnt], XmNsliderSize, editorHeight(exHistory->comEdit) / l); 
  cnt++;
  XtSetArg(args[cnt], XmNminimum, 0); cnt++;
  XtSetArg(args[cnt], XmNmaximum, editorHeight(exHistory->comEdit)); cnt++;

  XtSetValues(exHistory->scrollbar, args, cnt); /* adapt values of scrollbar */

  /******************** set window manager hints **************************/

  wm_hints.flags = IconPixmapHint | IconPositionHint;
  wm_hints.icon_pixmap =  XCreateBitmapFromData(display, 
	      XDefaultRootWindow(display), hist_bits, hist_width, hist_height);
  wm_hints.icon_x = 1060;
  wm_hints.icon_y = 500;

  fw = exFontWidth(getEditorFont(exHistory->comEdit));
  fh = exFontHeight(getEditorFont(exHistory->comEdit));
  histWidth = editorWidth(exHistory->numEdit) + scrlbWidth; 
  histHeight = editorHeight(exHistory->comEdit) +
                                       menuBarHeight(exHistory->menuBar);

  size_hints.flags = PMinSize | PResizeInc | PMaxSize;
  size_hints.min_width  = 15 * fw + histWidth;
  size_hints.min_height = 5 * fh + menuBarHeight(exHistory->menuBar);

  size_hints.max_width  = 2 * (histWidth + editorWidth(exHistory->comEdit));
  size_hints.max_height = 3 * histHeight;
  size_hints.height_inc = fh;
  size_hints.width_inc = fw;

  /******************** show widgets ***********************************/

  XtPopup(exHistory->histShell, XtGrabNone);

  XSetWMHints(display, XtWindow(exHistory->histShell), &wm_hints);
  XSetNormalHints(display, XtWindow(exHistory->histShell), &size_hints);

  setHistScrollbar(exHistory);

  setbuf(stdin, NULL);
  setbuf(stdout, NULL);

  dpy = ConnectionNumber(display);

  FD_ZERO(&readfds[0]);
  FD_SET(dpy, &readfds[0]);
  FD_SET(fileno(stdin), &readfds[0]);

  while (DELETED) {

    bcopy((char *) &readfds[0], (char *) &readfds[1], sizeof(readfds[0])); 
    XFlush(display);
    nfds = select(maxfd, &readfds[1], (fd_set *) 0, (fd_set *) 0, 
		  (struct timeval *) 0); 
    if (FD_ISSET(dpy, &readfds[1]))               /* process events */
      processEvents();    
    else if (FD_ISSET(fileno(stdin), &readfds[1]))  	  /* process IO */
      exHistoryio(exHistory);    
  }
}

/* getHistMenuResult ********************************************************
 * 
 ****************************************************************************/

static void getHistMenuResult(w, event, params, num_params)
     Widget w;
     XPropertyEvent *event;
     String *params;
     Cardinal *num_params;
{
  char *value;

  value = getMenuEventReturnValue(event);
  free(value);
}

/* processEvents *************************************************************
 * process all pending events 
 ****************************************************************************/

static processEvents()
{	
  static XEvent event;
  
  while (XtPending()) {
    XtNextEvent(&event);
    XtDispatchEvent(&event);
  }
}

/* HEdoScroll ***************************************************************
 * Callback procedure invoked whenever the user requests scrolling by
 * clicking over the scrollbar 
 ****************************************************************************/ 

static void HEdoScroll(widget, closure, callData)
     Widget widget;
     caddr_t closure;
     caddr_t callData;
{
  historyWindow *HEditor = (historyWindow *) closure;

  doScroll(widget, bufinfoptr(HEditor->comEdit), callData);
  doScroll(widget, bufinfoptr(HEditor->numEdit), callData);
  setHistScrollbar(HEditor);
}

/* exHistoryio ****************************************************************
 * inserts input string into editor buffer
 *****************************************************************************/

exHistoryio(hwindow)
     historyWindow *hwindow;
{
  char buf[512],
       *cmdstring,    /* holds null-terminated string read from command 
			 listener */  
       *soc,          /* points to 'start of command signal' in 'cmdstring' */ 
       *sigPtr,       /* used to search next EOC SOC signal */ 
       *curcom;       /* holds current command */ 
  int n;
  char *strpbrk(), *index();
  static char delim[] = {SOC, EOC, EOS};
  static SIGCOM = 0;       /* if set SOC signal was received but no EOC 
			     signal yet */

  if ((n = read(0, buf, 512)) > 0) {

    if (STOP == 1) 
      deleteComplCom(comStat);

    cmdstring = (char *) malloc(n + 1);
    strncpy(cmdstring, buf, n);
    cmdstring[n] = EOS;

    if (((sigPtr = strpbrk(cmdstring, delim)) != NULL) ||
	(SIGCOM == 1)) { 

      if (sigPtr && *sigPtr == EOC) {   /* 'end of command signal' received */ 
	SIGCOM = 0;
	if (STOP != 1) {
	  curcom = makeStr(cmdstring, sigPtr, strlen(cmdstring) 
			   - strlen(sigPtr) + 1);
	  compoundCom(comStat, curcom);
	  free(curcom);
	  addCommToHist(hwindow, comStat, comStat->complCom);
	}
	sigPtr = nextCommand(hwindow->comEdit, sigPtr);
      }
      else if (sigPtr && *sigPtr == SOC) {   /* 'start of command signal' 
					       received */ 
	if (SIGCOM == 1)
	  exError("HT", "end of command-signal missing !", NULL);
	SIGCOM = 1;
	sigPtr++;
      }
      else {
	if (SIGCOM == 1)                     /* pending SOC signal */ 
	  sigPtr = cmdstring;
      }

      while (sigPtr && *sigPtr != EOS) {
	soc = sigPtr;	      /* sigPtr points to 'start of command signal' */

	if (*sigPtr == EOS) {
	  SIGCOM = 1;                                 /* pending SOC signal */ 
	  return;
	}

	if ((sigPtr = index(soc, EOC)) == NULL) {                              
	  if (STOP != 1) 
	    compoundCom(comStat, soc);   /* save first part of next command */ 
	  SIGCOM = 1;                                 /* pending SOC signal */ 
	  return;
	}
	
	SIGCOM = 0;
	if (STOP != 1) {
	  curcom = makeStr(soc, sigPtr, strlen(soc) - strlen(sigPtr) + 1);
	  compoundCom(comStat, curcom);
	  free(curcom);
	  addCommToHist(hwindow, comStat, comStat->complCom);
	}
	sigPtr = nextCommand(hwindow->comEdit, sigPtr);
      }
    }

    else {    /* no EOC/SOC signal in received string or pending SOC signal */ 
      if (cmdstring)
	edit(bufinfoptr(hwindow->comEdit), cmdstring, strlen(cmdstring));
    }

    free(cmdstring);
  }
}

/* initHist ******************************************************************
 * initialize global variables of history editor and editor structure
 *****************************************************************************/

static void initHist()
{
  register i;

  comStat = (commandStatistics *) Malloc(sizeof(commandStatistics));
  comStat->comCount = 0;
  for (i = 0; i < MAXHIST; i++) {
    comStat->comSeq[i][0] = -1;
    comStat->comSeq[i][1] = -1;
  }

  MH_FUNCTIONS[0].func = (int (*)()) 0;
  MH_FUNCTIONS[0].funcName = (char *) 0;

  i = 1;
  MH_FUNCTIONS[i].func = (int (*)()) stopHistory;
  MH_FUNCTIONS[i].funcName = "stopHistory"; i++;
  MH_FUNCTIONS[i].func = (int (*)()) contHistory;
  MH_FUNCTIONS[i].funcName = "contHistory"; i++;
  MH_FUNCTIONS[i].func = (int (*)()) saveHistory;
  MH_FUNCTIONS[i].funcName = "saveHistory"; i++;
  MH_FUNCTIONS[i].func = (int (*)()) saveAsMacro;
  MH_FUNCTIONS[i].funcName = "saveAsMacro"; i++;
  MH_FUNCTIONS[i].func = (int (*)()) restoreHistory;
  MH_FUNCTIONS[i].funcName = "restoreHistory"; i++;
  MH_FUNCTIONS[i].func = (int (*)()) quitHistory;
  MH_FUNCTIONS[i].funcName = "quitHistory"; i++;
  MH_FUNCTIONS[i].func = (int (*)()) 0;
  MH_FUNCTIONS[i].funcName = NULL;
}

/* quitHistory **************************************************************
 * application function to quit historytracer and to make its windows invisible
 ***************************************************************************/

quitHistory(parms)
     RequestParm *parms;
{
  char msg[100];
  Arg args[5];
  Position x, y;

  if (! emptyBuffer(exHistory->comEdit)) {
    XtSetArg(args[0], XmNx, &x);
    XtSetArg(args[1], XmNy, &y);
    XtGetValues(exHistory->histShell, args, 2);
    switch (exNotify2User(exHistory->histShell, x + 20, y + 20, 
			  "save history?")) {
    case ANSWERYES:
      saveHistory((RequestParm *) NULL);
      break;
    case ANSWERCANCEL:
      return;
    default:
      break;
    }
  }
  
  XtUnmapWidget(exHistory->histShell);
  XtDestroyWidget(exHistory->histShell);

  deleteComplCom(comStat);

  exit(0);
}

/* stopHistory ***************************************************************
 * application function to stop taking down the history
 ****************************************************************************/

stopHistory(parms)
     RequestParm *parms;
{
  /* stop writing history */
  int actmask[5];
  
  STOP = 1;
  
  setMask(actmask, 5, 0);
  actmask[1] = 1;
  actmask[2] = 1;
  actmask[3] = 1;
  actmask[4] = 1;
  setInitialMask(mWins[0].menu, actmask);

  return(1);
}

/* contHistory ***************************************************************
 * application function to continue taking down the history
 ****************************************************************************/

contHistory(parms)
     RequestParm *parms;
{
  /* continue writing history */
  int actmask[5];
  
  STOP = 0;   /* input buffer loeschen ??? */

  setMask(actmask, 5, 0);
  actmask[0] = 1;
  actmask[2] = 1;
  actmask[3] = 1;
  actmask[4] = 1;
  setInitialMask(mWins[0].menu, actmask);

  return(1);
}

/* saveHistory ***************************************************************
 * application function to save a history sequence in a file
 ****************************************************************************/

saveHistory(parms)
     RequestParm *parms;
{
  int fd;
  register i;
  char *filename, *comStr, msg[200], ending[MAXFNAME];
  char *getenv();

  filename = exGetFilename(exHistory->histShell, -1, -1, "File name:", 
			   getenv("HOME"), MAXFNAME);
  
  if (! filename || ! *filename)
    return(-1);

  sprintf(ending, "%s.hist", filename);
  if (strlen(ending) > MAXFNAME) {
    editorMsg(exHistory->comEdit, "File name too long !");
/*    setError(ERR_MAX);*/
    return(-1);
  }

  if ((fd = open(ending, O_WRONLY | O_CREAT | O_TRUNC, MDS_UMASK)) == ERROR)
    {
      sprintf(msg, "Can't open -- %s -- !", filename);
      editorMsg(exHistory->comEdit, msg);
      setError(ERR_OPEN);
      free(filename);
      return(-1);
    }

  Write(fd, &(comStat->comCount), sizeof(int));
  for (i = 0; i < comStat->comCount; i++) 
    if (comStr = getRange(exHistory->comEdit, comStat->comSeq[i][0], 
			  comStat->comSeq[i][1], '\n')) {
      writeString(fd, comStr);
      free(comStr);
    }

  if (filename)
    free(filename);

  close(fd);
  return(1);
}

/* saveAsMacro ***************************************************************
 * application function to save a history sequence as macro in a file
 ****************************************************************************/

saveAsMacro(parms)
     RequestParm *parms;
{
  char *filename,                      /* name for new macro */
       *setname,                       /* macro set name */
       *comStr,                        /* holds contents of editor buffer */ 
       *start;                         /* beginning of macropath */
  char *getenv();

  if (!(start = strCopy(getenv("MACROPATH")))) {
    if (!(start = strCopy(getenv("HOME")))) {
      setError(ERR_NOEXIST);
      exAlert(exHistory->histShell, -1, -1, "Cannot locate your macro directory: \nSet the Environment variable MACROPATH!");
      return(-1);
    }
  }

  setname = exGetFilename(exHistory->histShell, -1, -1, 
			  "enter macro set name:", start, MAXFNAME);
  
  if (! setname || ! *setname)
    return(-1);

  filename = exGetFilename(exHistory->histShell, -1, -1, 
			   "enter macro name:", NULL, MAXFNAME);
  
  if (! filename || ! *filename)
    return(-1);

  comStr = getBufferContents(exHistory->comEdit);
  storeMacro(exHistory->histShell, setname, filename, comStr);

  return(1);
}

/* restoreHistory ************************************************************
 * application function to restore a history sequence from a file
 ****************************************************************************/

restoreHistory(parms)
     RequestParm *parms;
{
  int fd, i, count;
  char *filename, *comStr, msg[200];

  filename = exGetFilename(exHistory->histShell, -1, -1, "File name:", NULL, 
			   MAXFNAME);
  
  if (! *filename)
    return(-1);

  if (strlen(filename) > MAXFNAME) {
    editorMsg(exHistory->comEdit, "File name too long !");
    *filename = EOS;
  }

  if (access(filename,R_OK) == ERROR) {    /* no read permission */
    setError(ERR_READ);
    return(-1);
  }

  if ((fd = open(filename, O_RDONLY)) == ERROR) {
    sprintf(msg, "Cant't open -- %s -- !", filename);
    editorMsg(exHistory->comEdit, msg);
    setError(ERR_OPEN);
    return(-1);
  }

  Read(fd, &count, sizeof(int));

  for (i = 0; i < count; i++) {
    readString(fd, &comStr);
    if (comStr) {
      addCommToHist(exHistory, comStat, comStr);
      free(comStr);
    }
  }

  close(fd);
  return(1);
}
