/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
#include <stdio.h>
#include "exBasics.h"

#define TABSIZE 8
#define MAXLINES 512

static spaces;
static holecnt;
static char oline[2048];
static char *olp;
static char *holeptr[256];
static currentLine;
static char *FMTBUF[MAXLINES];
static offset;

static int putword();
static int putline();
static int justifyline();
static Boolean skipline();
static Boolean endofsentence();
static int freeBuffer();
static int copyBuffer();

static char *delimiters = " \t\n\r\f";

extern char *strchr(), *strtok(), *strCopy(), **Calloc();

/* exFormat: ****************************************************************
 * formats the text in the buffer according to the line-length given with the 
 * parameter length; input is the unformatted buffer (buffer), the number of 
 * lines in buffer (lines) and the requested line-length (length); The text 
 * will be formatted and stored into the allocated buffer result. The number 
 * of lines is returned.
 ****************************************************************************/

int exFormat(buffer, lines, maxlen, Offset, result, justify)
     char **buffer;     /* input buffer */
     int lines;		/* lines in buffer */
     int maxlen;	/* requested line length */
     int Offset;	/* offset in each line */
     char ***result;	/* the resulting buffer */
     Boolean justify;	/* if TRUE, lines are justified */
{
  char *ilp;
  char *p;
  int cnt;
  
  oline[0] = EOS;
  olp = oline;
  spaces = 0;
  holecnt = 0;
  currentLine = 0;	
  offset = Offset;
  
  for (cnt = 0; cnt < lines; cnt++) {
    ilp = strCopy(buffer[cnt]);
    if (*ilp == '\0') {			/* empty line */
      if (! skipline(1)) {
	freeBuffer(FMTBUF, currentLine);
	free(ilp);
	return(-1);
      }
    } else {				/* process line */
      p = strtok(ilp, delimiters);
      while (p) {
	putword(p, justify, maxlen);
	p = strtok(NULL, delimiters);
      }
      free(ilp);
    }
  }
  putline();
  copyBuffer(FMTBUF, result, currentLine);
  freeBuffer(FMTBUF, currentLine);
  return(currentLine);
} 

static putword(p, justify, maxlen)
     char *p;
     Boolean justify;
     int maxlen;
{
  int plen;
  
  plen = strlen(p);
  if ((olp - oline) + spaces + plen > maxlen) {
    if (justify && holecnt)
      justifyline(maxlen);
    putline();
  }
  if (spaces) {
    holeptr[++holecnt] = olp;
    for (  ;  (spaces > 0);  spaces--)
      *olp++ = ' ';
  }
  
  spaces = endofsentence(p, plen) ? 2 : 1;
  
  while (*p)
    *olp++ = *p++;
} 

static putline()
{
  *olp = '\0';
  if (*oline) {
    FMTBUF[currentLine] = (char *) malloc(offset + strlen(oline) + 1);
    sprintf(FMTBUF[currentLine], "%*s", offset + strlen(oline), oline);
    currentLine++;
  }
  *oline = '\0';
  olp = oline;
  spaces = 0;
  holecnt = 0;
}

static justifyline(maxlen)
     int maxlen;
{
  int n;
  register char *fp, *tp;
  static Boolean dir = False;
  
  if (dir)
    dir = False;
  else 
    dir = True;
  
  fp = olp - 1;
  olp = &oline[maxlen];
  tp = olp - 1;
  while (tp > fp) {
    while (fp >= holeptr[holecnt])
      *tp-- = *fp--;
    if (dir)
      n = ((tp - fp) - 1) / holecnt + 1;
    else
      n = (tp - fp) / holecnt;
    while (n--) 
      *tp-- = ' ';
    holecnt--;
  }
}

/* skipline: ********************************************************************
 * skip the specified number of lines; i.e. put empty lines into the resulting 
 * buffer and increment the line counter (currentLine) accordingly;
 *******************************************************************************/

static Boolean skipline(n)
     int n;
{
  putline();
  
  for (  ;  n > 0;  n--) {
    if (currentLine >= MAXLINES)
      return(False);
    FMTBUF[currentLine++] = strCopy(" ");
  }
  return(True);
}

static Boolean endofsentence(p, plen)
     char *p;
     int plen;
{
  if (plen < 3)
    return(False);
  if (!strchr(".:?!;,()[]{}|<>", *(p + plen - 1)))
    return(False);
  return(True);
}

static freeBuffer(buffer, lines)
     char **buffer;
     int lines;
{
  register i;
  for (i = 0; i < lines; i++)
    free(buffer[i]);
}

static copyBuffer(from, to, lines)
     char **from, ***to;
     int lines;
{
  register i;
  
  *to = Calloc(lines, sizeof(char *));
    
  for (i = 0; i < lines; i++) 
    (*to)[i] = strCopy(from[i]);
}
