/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
#include <sys/types.h>
#include <exBasics.h>
#include <exEdit.h>
#include <exLayout.h>
#include <exMDS.h>
#include <exMenu.h>
#include "exEntry.h"
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <Xm/DrawingA.h>
#include <Xm/LabelG.h>
#include <Xm/BulletinB.h>

static char SccsID[] = "@(#)exEntry.c	1.2\t11/10/92";            /* sccs keyword */

BufInfo *bufinfoptr();

static Display *display;
static XContext entryCtxt;
static GC gc;

static void initEntry();
static void exGrabEntry();
static flashEntry();
static closeEntry();
static abortEntry();
static expandBuffer();
static showExpansionList();
static entryHelp();
static void exEntryExposeProc();
static exEntry *exCreateEntry();

static void exEntryKeyTestProc();

static XtTranslations ptt = NULL;

static char entry_translations[] = 
  " <Expose>	: exEntryExposeProc()";

static XtActionsRec entry_actions[] = {
  { "exEntryExposeProc", exEntryExposeProc }
};

static editCmd entryCmdTable [] = {
  {ctrl('a'), beglin},		/* beginning of line */
  {ctrl('b'), left},		/* cursor left  */
  {ctrl('c'), abortEntry},      /* close entry and clear buffer */
  {ctrl('d'), delnxt},	       	/* delete next character */
  {ctrl('e'), endlin},		/* end of line */
  {ctrl('f'), right},		/* cursor right */
  {ctrl('j'), closeEntry},	/* close this entry */
  {ctrl('k'), krest},		/* delete to end of line */
  {ctrl('l'), newscr},		/* redraw screen */
  {ctrl('m'), closeEntry},	/* close this entry */
  {ctrl('n'), dwncol},		/* down in column */
  {ctrl('o'), openln},	        /* open line above */
  {ctrl('p'), upcol},		/* up in column */
  {ctrl('v'), dwnpage},		/* forward page */
  {127,       delprev},         /* delete previous char */
  {EOS,       (int (*)()) 0},	/* end of list */
};
	
static editCmd getFileCmdTable [] = {
  {ctrl('a'), beglin},		/* beginning of line */
  {ctrl('b'), left},		/* cursor left  */
  {ctrl('c'), abortEntry},      /* close entry and clear buffer */
  {ctrl('d'), delnxt},	       	/* delete next character */
  {ctrl('e'), endlin},		/* end of line */
  {ctrl('f'), right},		/* cursor right */
  {ctrl('i'), expandBuffer}, 	/* expand the buffer contents to a file name */
  {ctrl('j'), closeEntry},	/* close entry */
  {ctrl('k'), krest},		/* delete to end of line */
  {ctrl('l'), newscr},		/* redraw screen */
  {ctrl('m'), closeEntry},	/* close entry */
  {ctrl('n'), dwncol},		/* down in column */
  {ctrl('p'), upcol},		/* up in column */
  {ctrl('u'), showExpansionList}, /* show the list of possible expansions */
  {ctrl('v'), dwnpage},		/* forward page */
  {127,       delprev},         /* delete previous char */
  {EOS,     (int (*)()) 0},	/* end of list */
};		

static editCmd ctrlxTable [] = {		       
  {'k', killBuffer},		/* clear the current buffer */
  {'o', newline},		/* insert newline */	       
  {'?', entryHelp},		/* display help information */
  {ctrl('c'), abortEntry},      /* close entry and clear buffer */
  {EOS, (int (*)()) 0},		
};

static editCmd escTable [] = {
  {'=', memory},		/* display statistics */
  {'b', wordl},			/* move to left word */
  {'f', wordr},			/* move to right word */
  {'w', fdelword},		/* delete next word */
  {127, bdelword},		/* delete previous word */
  {ctrl('c'), abortEntry},        /* close entry and clear buffer */
  {EOS, (int (*)()) 0},		
};

static cmdDescr entryCmds[] = {
  {closeEntry,        " close entry"},
  {abortEntry,        " close entry, kill buffer"},
  {entryHelp,	      " provide some help information"}
};

static cmdDescr getFileCmds[] = {
  {closeEntry,        " close entry"},
  {abortEntry,        " close entry, clear buffer"},
  {expandBuffer,      " expand into a file name"},
  {showExpansionList, " show the possible expansions in a menu"},
};

static EditorDescr entryDescr = {
  (int (*)()) 0,			/* input function */
  (char *) 0,				/* initial file */
  "?entry?",				/* editor name */
  (char **) 0,				/* external buffer */
  (Widget) 0,				/* msg window */
  1,    				/* number of lines in buffer */
  1,					/* lines in window */
  0,					/* cols  in window */
  0, 					/* width of window */
  0,					/* height of window */
  0                       		/* flags */
};

/* exMakeEntry: ***************************************************************
 * create the entry data-structures, the underlying editor and initialize
 * the command tables for the editor; return a pointer to the structure.
 *****************************************************************************/

exEntry *exMakeEntry(parent, buf, buflin, size, label, 
		     cmdTable, ecTable, cxTable, addCmds, cursorOn, hlp)
     Widget parent;        	/* Parent in which to create entry window */
     char **buf;		/* holds the default or the inserted string */
     int buflin;                /* bufferlines of editor */
     int size;			/* the size of the buffer */
     char *label;		/* the entry label */
     editCmd *cmdTable,         /* command table for entry editor */
             *ecTable, *cxTable;/* esc-command / ctrl-x-command tables */ 
     cmdDescr *addCmds;
     Boolean cursorOn;          /* if True, entry is active when shown */
     Boolean hlp;               /* if False, no help file is created */
{
  exEntry *entry;

  entry = exCreateEntry(parent, buf, buflin, size, label, 
			cursorOn, hlp);
  initCmdTables(entry->entryEditor, 
		cmdTable ? cmdTable : entryCmdTable, 
		ecTable ? ecTable : escTable, 
		cxTable ? cxTable : ctrlxTable, 
		addCmds ? addCmds : entryCmds);

  entry->flag = MAPPED;
  return(entry);
}

/* exCreateEntry: ************************************************************
 * create the entry data-structures and the underlying editor and return
 * a pointer to the structure.
 ****************************************************************************/

static exEntry *exCreateEntry(parent, buf, buflin, size, label, 
			      cursorOn, hlp)
     Widget parent;        	/* entry shell in which to create entry */
     char **buf;		/* holds the default or the inserted string */
     int buflin;                /* bufferlines of editor */
     int size;			/* the size of the buffer */
     char *label;		/* the entry label */
     Boolean cursorOn;          /* if true, entry is active when shown */
     Boolean hlp;               /* if false, no help file is created */
{
  exEntry *entry;
  Arg args[10];
  int cnt;
  XmFontList fl;

  initEntry(parent);

  if (buf) {
    entryDescr.flags = ED_EXTERNBUFFER;
    entryDescr.buffer = buf;
  } else {
    entryDescr.flags = 0;
    entryDescr.buffer = NULL;
  }

  if (buflin > 0)
    entryDescr.bufferLines = buflin;
  entryDescr.cols = size;

  entry = (exEntry *) Malloc(sizeof(exEntry));
  entry->shell = parent;
  entry->bufSize = size;

  /* create the entry's main widget */
  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, entry); cnt++;
  /*XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
  XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;*/
  XtSetArg(args[cnt], XmNtopShadowPixmap, exWhitePM(XtDisplay(parent))); cnt++;
  XtSetArg(args[cnt], XmNbottomShadowPixmap, 
	   		exWhitePM(XtDisplay(parent))); cnt++;

  entry->entryWindow = XtCreateManagedWidget("entry", 
			xmFormWidgetClass, entry->shell, args, cnt);
  
  /* create the label widget */
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNlabelString,  
	   XmStringCreate(label, XmSTRING_DEFAULT_CHARSET)); cnt++;

  entry->label = XtCreateManagedWidget("label", xmLabelWidgetClass, 
				      entry->entryWindow, args, cnt);

  cnt = 0;
  XtSetArg(args[cnt], XmNfontList, &fl); cnt++;
  XtGetValues(entry->label, args, cnt);

  if (fl)
    entry->f = getFont(fl);
  else
    entry->f = exEditorFont(XtDisplay(parent));

  /* create the editor window */
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, entry->label); cnt++;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  
  if (hlp)
    entryDescr.flags |= ED_HELP;
  if (cursorOn)
    entryDescr.flags |= ED_ACTIVE;
   entryDescr.flags |= ED_ENTRY;

  /* make the editor instance */
  entry->entryEditor = 
    makeEditor(entry->entryWindow, "editor", &entryDescr, args, cnt);

  XSaveContext(display, (Window) entry->entryEditor, entryCtxt, 
	       (caddr_t) entry);
  return(entry);
}

/* destroyEntry: **************************************************************
 * called to destroy the entry and free its data structures 
 *****************************************************************************/

destroyEntry(entry)
     exEntry *entry;
{
  XtDestroyWidget(entry->shell);
  XDeleteContext(display, entry->entryEditor, entryCtxt);
  free((exEntry *)entry);
}

/* exGrabentry: *************************************************************
 * grabs or ungrabs  the keyboard for the specified entry's editor window;
 ***************************************************************************/

static void exGrabEntry(entry)
     exEntry *entry;
{
  BufInfo *bf;
  Display *d = XtDisplay(entry->shell);
  
  bf = bufinfoptr(entry->entryEditor);

  exGrabAll(bf->edt, None);

  XWarpPointer(d, None, XtWindow(bf->edt), 0, 0, 0, 0, 0, 
	       exFontHeight(entry->f) / 2);
}
    
/* closeEntry: ****************************************************************
 * is called whenever the user wants to close the entry.
 * The widget is popped down. 
 *****************************************************************************/

static closeEntry(editor)
     Editor editor;
{
  exEntry *entry;
  BufInfo *bufinfoptr();
  
  if (! XFindContext(display, (Window) editor, entryCtxt,(caddr_t *) &entry)) {
    entry->flag &= ~MAPPED;
    entry->flag &= ~GRABBED;

    exUngrabAll(bufinfoptr(entry->entryEditor)->edt);
    
    XtPopdown(entry->shell);
  }
}

/* abortEntry: ****************************************************************
 * is called whenever the user wants to abort the entry, The buffer is cleared 
 * and the window is unmapped. 
 *****************************************************************************/

static abortEntry(editor)
     Editor editor;
{
  exEntry *entry;
  BufInfo *bufinfoptr();
  
  if (!  XFindContext(display, (Window) editor, entryCtxt,(caddr_t *)&entry)) {
    clearBuffer(entry->entryEditor);
    entry->flag &= ~MAPPED;
    entry->flag &= ~GRABBED;

    exUngrabAll(bufinfoptr(entry->entryEditor)->edt);

    XtPopdown(entry->shell);
  }
}

/* expandBuffer: *********************************************************
 * starting at <path> the patterns pointed to by <pattern> are expanded. The 
 * result is stored into buffer; the number of successful matches is returned.
 *****************************************************************************/

static expandBuffer(editor)
     Editor editor;
{
  exEntry *entry;
  char *word;
  exExpandInfo *info;
  char *getcwd();

  if (XFindContext(display, (Window) editor, entryCtxt, (caddr_t *) &entry))
    return;

  if ((word = getBufferContents(editor)) == NULL)
    return;
  
  if (word[0] )
    word[strlen(word) - 1] = EOS;		/* delete newline */

  if (! (info = setExpandInfo(entry->shell, word, (int (*)()) getcwd))) {
    free(word);
    return;
  }
  
  if (! expandFileName(info, False, False)) {	/* expansion failed */
    flashEntry(entry);
    if (info->retBuf)
      free(info->retBuf);
    free(word);
    return;
  }

  editorFillBuffer(editor, info->retBuf, True);

  if (info->matchCnt > 1) 			/* more than one match */
    flashEntry(entry);

  free(word);
  if (info->retBuf)
    free(info->retBuf);
}

/* showExpansionList: *********************************************************
 * show the user a menu of possible expansions and let him select 
 *****************************************************************************/

static showExpansionList(editor)
     Editor editor;
{
  exEntry *entry;
  char *word;
  exExpandInfo *info;
  char *getcwd();

  if (XFindContext(display, (Window) editor, entryCtxt, (caddr_t *) &entry))
    return(-1);

  if ((word = getBufferContents(editor)) == NULL)
    return(-1);
  
  if (word[0] )
    word[strlen(word) - 1] = EOS;		/* delete newline */

  if (! (info = setExpandInfo(entry->shell, word, (int (*)()) getcwd))) {
    free(word);
    return(-1);
  }
  
  if (! expandFileName(info, False, True)) {	/* expansion failed */
    flashEntry(entry);
    if (info->retBuf)
      free(info->retBuf);
    free(word);
    return(-1);
  }
  
  if (info->retBuf) {
    editorFillBuffer(editor, info->retBuf, True);
    free(info->retBuf);
  }

  free(word);
  return(0);
}

/* flashEntry: ****************************************************************
 * make the entry flash; is used when an error has ocurred.
 *****************************************************************************/

static flashEntry(entry)
     exEntry *entry;
{
  Display *d = XtDisplay(entry->entryWindow);
  Window w = XtWindow(entry->shell);

  XSetWindowBorderPixmap(d, w, exWhitePM(d));
  XSync(d, 0);
  exWait(10);
  XSetWindowBorderPixmap(d, w, exBlackPM(d));
  XSync(d, 0);
  exWait(5);
  XSetWindowBorderPixmap(d, w, exWhitePM(d));
  XSync(d, 0);
  exWait(5);
  XSetWindowBorderPixmap(d, w, exBlackPM(d));
  XSync(d, 0);
}

/* entryHelp: ****************************************************************
 * show the help editor. Special care has to be taken because when the help 
 * editor comes up the focus is set on the entry editor, so that all keys are 
 * forwarded to the entry editor. Thus, we reset the focus until the help
 * editor is closed. 
 *****************************************************************************/

static entryHelp(editor)
     Editor editor;
{
  help(editor);
}

/* initEntry: *************************************************************
 * initialize context of entry
 ***************************************************************************/

static void initEntry(wd)
     Widget wd;
{
  static Boolean initialised = False;
  Display *d = XtDisplay(wd);
  Window w = XDefaultRootWindow(d);

  if (initialised)
    return;

  entryCtxt = XUniqueContext();
  display = d;
  
  gc = XCreateGC(d, w, 0, 0);
  XSetBackground(d, gc, exWhitePixel(d));
  XSetForeground(d, gc, exBlackPixel(d));
  XSetFunction(d, gc, GXcopy);
  
  initialised = True;
}
  
/*****************************************************************************
 *
 *                exGetFileName, exGetEntry
 *
 *****************************************************************************/

/* exGetFilename: *************************************************************
 * ask the user for a file name (provide file-name completion)
 ***************************************************************************/

char *exGetFilename(parent, x, y, msg, buf, len)
     Widget parent;
     int x, y;
     char *msg;
     char *buf;
     int len;
{
  char *filename, **ptr; 
  exEntry *entry;
  XEvent event;
  Widget entryShell;
  Arg args[5];
  int cnt;
  Dimension height, width, w, h;

  if (buf) {
    filename = strCopy(buf);
    ptr = &filename;
  } else {
    filename = NULL;
    ptr = NULL;
  }

  cnt = 0;
  XtSetArg(args[cnt], XmNkeyboardFocusPolicy, XmPOINTER); cnt++;

  entryShell = XtCreatePopupShell("entry shell", transientShellWidgetClass,
				  parent, args, cnt);

  if (! (entry = exMakeEntry(entryShell, ptr, 1, MIN(len, MAXFNAME), 
			     msg, getFileCmdTable, escTable, ctrlxTable, 
			     getFileCmds, True, False)))
    return(NULL);


  /* dimensions of label */
  cnt = 0;
  XtSetArg(args[cnt], XmNwidth, &w); cnt++;
  XtSetArg(args[cnt], XmNheight, &h); cnt++;
  XtGetValues(entry->label, args, cnt);
  
  /* dimensions of editor */
  cnt = 0;
  XtSetArg(args[cnt], XmNwidth, &width); cnt++;
  XtSetArg(args[cnt], XmNheight, &height); cnt++;
  XtGetValues(bufinfoptr(entry->entryEditor)->w, args, cnt);

  w += width;
  h = MAX(h, height);

  if (!ptt) {
    ptt = XtParseTranslationTable(entry_translations);
    XtAddActions(entry_actions, 1);
  }

  XtOverrideTranslations(entry->entryWindow, ptt);

  exWidgetPos(entry->shell, x, y, width, height);

  XtPopup(entry->shell, XtGrabExclusive);

  while (entry->flag & MAPPED) {
    XtNextEvent(&event);
    XtDispatchEvent(&event);
  }
  
  filename = getBufferContents(entry->entryEditor);

  if (filename && *filename)
    filename[strlen(filename) -1 ] = EOS;	/* get rid of <nl> */

  destroyEntry(entry);

  return(filename);
}

/* exGetEntry: *************************************************************
 * ask the user
 ***************************************************************************/

char *exGetEntry(parent, x, y, msg, buf, len)
     Widget parent;
     int x, y;
     char *msg;
     char *buf;				/* buffer if provided */
     int len;				/* size of buffer */
{
  char *buffer, **ptr;
  exEntry *entry;
  XEvent event;
  Widget entryShell;
  Arg args[5];
  int cnt;
  Dimension width, height, w, h;
  
  if (buf) {                                   /* default string */
    buffer = strCopy(buf); 
    ptr = &buffer;
  } else {
    buffer = NULL;
    ptr = NULL;
  }

  cnt = 0;
  XtSetArg(args[cnt], XmNborderWidth, 3); cnt++;
  XtSetArg(args[cnt], XmNkeyboardFocusPolicy, XmPOINTER); cnt++;

  entryShell = XtCreatePopupShell("entry shell", transientShellWidgetClass,
				  parent, args, cnt);

  if (! (entry = exMakeEntry(entryShell, ptr, 1, MIN(len, EX_MAXSTR), 
			     msg, entryCmdTable, escTable, ctrlxTable,  
			     entryCmds, True, False)))
    return(NULL);

  /* dimensions of label */
  cnt = 0;
  XtSetArg(args[cnt], XmNwidth, &w); cnt++;
  XtSetArg(args[cnt], XmNheight, &h); cnt++;
  XtGetValues(entry->label, args, cnt);
  
  /* dimensions of editor */
  cnt = 0;
  XtSetArg(args[cnt], XmNwidth, &width); cnt++;
  XtSetArg(args[cnt], XmNheight, &height); cnt++;
  XtGetValues(bufinfoptr(entry->entryEditor)->w, args, cnt);

  w += width;
  h = MAX(h, height);

  if (!ptt) {
    ptt = XtParseTranslationTable(entry_translations);
    XtAddActions(entry_actions, 1);
  }

  XtOverrideTranslations(entry->entryWindow, ptt);

  exWidgetPos(entry->shell, x, y, width, height);

  XtPopup(entry->shell, XtGrabExclusive);
  
  while (entry->flag & MAPPED) {
    XtNextEvent(&event);
    XtDispatchEvent(&event);
  }
  
  buffer = getBufferContents(entry->entryEditor);

  if (buffer && *buffer)
    buffer[strlen(buffer) -1 ] = EOS;	/* get rid of <nl> */

  destroyEntry(entry);

  return(buffer);
}

/* exEntryExposeProc: ********************************************************
 * callback function that is called when the entry is exposed; the function
 * grabs the pointer with respect to the corresponding entry.
 ****************************************************************************/

/*ARGSUSED*/
static void exEntryExposeProc(w, event, params, num_params)
     Widget w;
     XExposeEvent *event;
     String *params;
     Cardinal *num_params;
{
  Arg args[2];
  exEntry *entry;

  XtSetArg(args[0], XmNuserData, &entry);
  XtGetValues(w, args, 1);

  if (! (entry->flag & GRABBED)) {
    entry->flag |= GRABBED;
    exGrabEntry(entry);
 }
}
