#ifndef _ex_exMenuP_h_
#define _ex_exMenuP_h_

/*  Standard C headers  */
#include <stdio.h>
#include <sys/signal.h>

/*  X headers  */
#include <X11/Xatom.h>
#include <X11/IntrinsicP.h>
#include <X11/Shell.h>

/*  Xm headers  */
#include <Xm/Xm.h>
#include <Xm/BulletinB.h>
#include <Xm/DrawingA.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>
#include <Xm/ScrollBar.h>
#include <Xm/ScrollBarP.h>
#include <Xm/Separator.h>

/* exMenu specific headers */
#include <exBasics.h>
#include <exLayout.h>
#include "exMDS.h"

#define EX_MENU_RETURN_VALUE_ATOM_NAME  "__exMenuReturnValue"
#define EX_MENU_TOGGLE_BUTTON_ATOM_NAME "__exMenuToggleButton"
#define MH_INITIAL_DELAY  50		/* delay for repeated scrolling */

/* menu specific flags */
#define MH_ACTIVE       0x1   	/* selection state of an item changed */ 
#define SCROLLABLE      0x2    	/* list is scrollable */
#define POPUP           0x4     /* popup menu */  
#define PULLDOWN        0x8     /* pulldown menu */
#define PANEL           0x10    /* panel */
#define NESTED_ITEM     0x20	/* entry has requests specified */
#define MENU_MAPPED     0x40	/* menu-pane has is already mapped */
#define ARROW_SELECTED  0x80    /* button has been released in arrow region */
#define ARROW_REGION    0x100   /* button has entered arrow region */
#define MH_NOEXEC       0x200   /* do not allow execution for this menu */
#define MH_FIRST_SCRL   0x400   /* set at start of continous scrolling */
#define MH_ABORT        0x800   /* user aborted a menu selection */
#define MH_INITVALUE    0x1000	/* initial values for item defined */
#define FIXED_WIDTH     0x2000  /* fixed width for menu pane */
#define ARROW_OR_TICK   0x4000  /* menus has item(s) with arow or tick */
#define MULTI_SEL	0x8000  /* a multi selection menu */
#define MH_FIRST_MAP    0x10000 /* menu is mapped the first time */

/* values returned by findItem */
#define INTITLE        -2	/* cursor in title */
#define INDONEBOX      -3 	/* cursor in ok box */
#define OUTSIDE        -4       /* cursor outside of menu */
#define SB_INSIDE      -5	/* inside scrollbar */

/* constants used in reqViaMenu() to position nested menus */
#define MH_X_OFFSET    10	/* x offset for nested menus */
#define MH_Y_OFFSET    10	/* y offset for nested menus */

/* constants defined to determine the menu layout */
#define MH_BORDER      1        /* border width of menus */
#define MH_MARGIN      4	/* margin width used in many situations */
#define MH_TITLEGAP    10       /* gap between title and first item */
#define MH_PANEGAP     5        /* gap between pulldown-button and menu pane */
#define MH_TITLE       2	/* gap between upper border and title */
#define MH_SEPWIDTH    4        /* line between items and done-button */
#define MH_MENUBUTTON  20	/* max. number of buttons in menubar */
#define MH_SCRLBARGAP  2	/* gap between scrollbar and menu items */
#define ITEM_INDENT    2	/* space in front of item label */
#define ITEM_GAP       6	/* space between key & item, item & arrow */
#define COL_GAP        8        /* gap between cols of item */

#define DOWN           0     	/* scroll downwards */
#define UP	       1       	/* scroll upwards */

typedef struct {
  String name;				/* item name */
  Position xpos, ypos;			/* pos of entry-string in menu pane */
  Dimension labelOffset;		/* pos. of arrow for nested menus */
  Dimension width;			/* width of this item */
  Position ppos;			/* y-pos of pixmap */
  char key;				/* accelerator key */
  unsigned long flags;			/* item specific flags */
  Dimension height;			/* height of current item */
  int baseLine;				/* ascent of font */
  String i_return;			/* the return pattern */
  Cardinal reqCnt;			/* number of requestys */
  MHiReq *requests[MDS_MAXREQ];		/* table of requests */
  act_mask activityMask;		/* item specific mask */
  String hlpMsg;			/* help message */
  int (* hlpFnct)();			/* help function */
} menuEntry;

typedef struct _exMenu {
  Widget w;				/* the popup-shell */
  Widget shape;				/* top-level drawing area */
  Widget form;				/* top-level menu form */
  Widget title;				/* title widget (pop-up menus) */
  Widget title_sep;			/* separator: title and pane */
  Widget pane;				/* the pane widget */
  Widget done_sep;			/* separator: pane and done box */
  Widget done_box;			/* done box (multiselection menu) */
  Widget button;			/* button widget (pulldown menu) */
  Widget scrollbar;			/* scrollbar widget */
  struct _exMenubar *menubar;		/* associated menubar (pulldown menu)*/
  String name;				/* the menu title */
  String applName;			/* name of the application node */
  MHmenu *mds_structure;		/* points to the MDS structure */
  XtIntervalId sb_timer;		/* for continous scrolling */
  unsigned long flags;			/* flags */
  int itemsPerCol;			/* items in each col */
  int cols;				/* no. of cols for panels */
  Dimension colWidth;			/* width of each col */
  int dir;				/* to remember the last scroll dir */
  Pos pos;				/* position information */
  Position lm;				/* x-pos. leftmost item */
  Position x, y;			/* abs. pos of menu pane */
  int yb;				/* position of done - box */
  int xs, ys;				/* pos. of scrollbar widget */
  int baseLine;				/* ascent of font */
  Dimension hs;				/* height of scrollbar */
  int slider_pos;
  Dimension menuWidth;			/* height and ... */
  Dimension menuHeight;			/* ... width of menu */
  Dimension titleHeight;		/* height of title: popup menu */
  Dimension sepHeight;			/* separator height */
  Dimension doneHeight;			/* height of done box: multisel menu */
  Dimension titleFontHeight;		/* height of title font */
  Dimension sb_width;			/* width of scrollbar */
  char key;				/* accelerator key */
  XFontStruct *itemFont;		/* font to display items */
  XFontStruct *titleFont;		/* font for title and done box */
  int itemCnt;				/* number of entries */
  int itemDisplayCnt;        		/* items displayed in menu */
  int topItem;				/* index of top item displayed */
  Boolean erase;			/* erase initial values ? */
  Boolean mapped;			/* TRUE when widget mapped */
  int selCnt;				/* number of selectable items */
  MDS_CNT_TYPE selType;                 /* one of <, >, <=, >=, = */
  int selItemCnt;			/* the number of selected items */
  int selections[MDS_MAXITEM];		/* indices of selected items */
  String returnValues[MDS_MAXITEM];     /* values of selected items */
  String assembledReturnValue;          /* returnValue as returned */
  act_mask activityMask;		/* current activity mask */
  act_mask initialMask;			/* initial activity mask */
  menuEntry *entries[MDS_MAXITEM]; 	/* table of entries */
  int prev;				/* previous selected item */
  String hlpMsg;			/* help message */
  int (* hlpFnct)();			/* help function */
  XtTranslations ptt;			/* compiled translations */
} exMenu;

typedef struct {
  exMenu *menu;				/* pointer to menu structure */
  Boolean active;			/* TRUE, if the menu is active */
  Widget button;			/* the button widget */
  XtTranslations ptt;			/* compiled translaytions */
} exButton;

typedef struct _exMenubar {
  Widget menubar;
  int num_buttons;
  exButton buttons[MH_MENUBUTTON];
  XtTranslations ptt;			/* compiled translations */
  int flags;
} exMenubar;

typedef struct {
  exMenu *menu;
  char *name;
  MDS_MENU_TYPE type;	
} MenuDescr;

#endif
